# -*- mode: python; coding: utf-8 -*-
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from unittest import mock
import pytest  # noqa # caplog

from f3wikidatabot.bot import Bot
from f3wikidatabot.tests.wikidata import WikidataHelper


class TestBot(object):
    def setup_class(self):
        self.site = WikidataHelper().login()

    def teardown_class(self):
        self.site.logout()

    def test_factory(self):
        Bot.factory(["--verbose"])
        assert logging.getLogger("f3wikidatabot").getEffectiveLevel() == logging.DEBUG

        b = Bot.factory([])
        assert logging.getLogger("f3wikidatabot").getEffectiveLevel() == logging.INFO

        assert len(b.plugins) > 0

        plugin = "License"
        b = Bot.factory(["--verbose", "--plugin=" + plugin])
        assert 1 == len(b.plugins)
        assert plugin == b.plugins[0].__class__.__name__

        b = Bot.factory(
            [
                "--verbose",
                "--plugin=License",
            ]
        )
        assert 1 == len(b.plugins)

    @mock.patch.object(Bot, "run_items")
    @mock.patch.object(Bot, "run_query")
    def test_run(self, m_query, m_items):
        b = Bot.factory([])
        b.run()
        m_query.assert_called_with()
        m_items.assert_not_called()

        m_query.reset_mock()
        m_items.reset_mock()
        b = Bot.factory(["--verbose", "--item=Q1"])
        b.run()
        m_items.assert_called_with()
        m_query.assert_not_called()

    @mock.patch("f3wikidatabot.license.License.run")
    def test_run_items(self, m_run):
        b = Bot.factory(
            [
                "--verbose",
                "--item=Q1",
                "--plugin=License",
            ]
        )
        b.run()
        m_run.assert_called_with(mock.ANY)

    @mock.patch("f3wikidatabot.license.License.run")
    @mock.patch("pywikibot.pagegenerators.WikidataSPARQLPageGenerator")
    def test_run_query_default(self, m_query, m_run):
        b = Bot.factory(
            [
                "--verbose",
                "--plugin=License",
            ]
        )
        m_query.side_effect = "one page"
        b.run()
        m_run.assert_called_with(mock.ANY)

    @mock.patch("f3wikidatabot.license.License.run")
    @mock.patch("pywikibot.pagegenerators.WikidataSPARQLPageGenerator")
    def test_run_query_items(self, m_query, m_run, caplog):
        b = Bot.factory(
            [
                "--verbose",
                "--filter=license-verify",
                "--plugin=License",
            ]
        )
        m_query.side_effect = "one page"
        b.run()

        for record in caplog.records:
            if "running query" in record.message:
                assert "?license" in record.message



# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import logging
import re

import pywikibot
from pywikibot import pagegenerators as pg

from f3wikidatabot import plugin

log = logging.getLogger(__name__)


class License(plugin.Plugin):
    def __init__(self, *args):
        super(License, self).__init__(*args)
        self.license2item = None
        self.licenses = None

    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument(
            "--license",
            action="append",
            default=[],
            help="only consider this license (can be repeated)",
        )
        return parser

    @staticmethod
    def filter_names():
        return ["license-verify", "no-license"]

    def get_query(self, filter):
        format_args = {
            "license": self.P_copyright_license,
            "subclass_of": self.P_subclass_of,
            "instance_of": self.P_instance_of,
            "open_source": self.Q_open_source_license.getID(),
            "free_software": self.Q_free_software_license.getID(),
            "retrieved": self.P_retrieved,
            "delay": self.args.verification_delay,
        }
        if filter == "license-verify":
            query = """
            SELECT DISTINCT ?item WHERE {{
              {{
                ?item p:{license} ?license .
                ?license ps:{license}/wdt:{instance_of}?/wdt:{subclass_of}*
                    wd:{open_source}.
              }} Union {{
                ?item p:{license} ?license .
                ?license ps:{license}/wdt:{instance_of}?/wdt:{subclass_of}*
                    wd:{free_software}.
              }}
              OPTIONAL {{
                 ?license prov:wasDerivedFrom/
                     <http://www.wikidata.org/prop/reference/{retrieved}>
                     ?retrieved
              }}
              FILTER (!BOUND(?retrieved) ||
                      ?retrieved < (now() - "P{delay}D"^^xsd:duration))
            }} ORDER BY ?item
            """.format(
                **format_args
            )
        elif filter == "no-license":
            format_args.update(
                {
                    "foss": self.Q_free_and_open_source_software.getID(),
                    "free_software": self.Q_free_software.getID(),
                    "open_source_software": self.Q_open_source_software.getID(),
                    "public_domain": self.Q_public_domain.getID(),
                    "software": self.Q_software.getID(),
                }
            )
            query = """
            SELECT DISTINCT ?item WHERE {{
               {{
                 ?item p:{instance_of}/ps:{instance_of}/wdt:{subclass_of}*
                    wd:{foss}.
               }} Union {{
                 ?item p:{instance_of}/ps:{instance_of}/wdt:{subclass_of}*
                    wd:{free_software}.
               }} Union {{
                 ?item p:{instance_of}/ps:{instance_of}/wdt:{subclass_of}*
                    wd:{open_source_software}.
               }} Union {{
                 ?item p:{instance_of}/ps:{instance_of}/wdt:{subclass_of}*
                    wd:{public_domain}.
                 ?item p:{instance_of}/ps:{instance_of}/wdt:{subclass_of}*
                    wd:{software}.
               }}
               FILTER NOT EXISTS {{ ?item p:{license} ?license }}
            }} ORDER BY ?item
            """.format(
                **format_args
            )
        else:
            query = None
        return query

    def run(self, item):
        self.fixup(item)
        self.verify(item)

    def verify(self, item):
        item.get()
        self.debug(item, " verifying")

    def set_en_licenses(self):
        self.licenses = {
            "en": {
                "names": dict([(l, l) for (l, i) in self.license2item.items()]),
            },
        }

    def get_names(self, lang):
        if self.license2item is None:
            self.set_license2item()
            self.set_en_licenses()
            self.set_redirects("en")
        if lang not in self.licenses:
            self.set_names(lang)
            self.set_redirects(lang)
        licenses = self.licenses[lang]
        return list(licenses["names"].keys()) + list(licenses["redirects"].keys())

    def set_redirects(self, lang):
        redirects = {}
        for name in self.licenses[lang]["names"].keys():
            redirects[name] = name
            for redirect in self.get_redirects(name, lang):
                redirects[redirect] = name
        self.licenses[lang]["redirects"] = redirects

    def set_names(self, lang):
        lang2en = {}
        self.licenses[lang] = {"names": lang2en}
        for english in self.licenses["en"]["names"].keys():
            title = self.translate_title(english, lang)
            if title is not None:
                log.debug("set_names " + lang + " " + english + " => " + title)
                lang2en[title] = english

    def get_item(self, license, lang):
        licenses = self.licenses[lang]
        canonical = licenses["redirects"][license]
        english = licenses["names"][canonical]
        return self.license2item[english]

    def set_dbname2item(self):
        query = """
            SELECT DISTINCT ?item WHERE {{
              ?item wdt:{dbname} ?dbname.
        }}
        """.format(
            dbname=self.P_Wikimedia_database_name
        )
        log.debug("set_dbname2item " + query)
        self.license2item = {}
        enwiki = pywikibot.site.APISite.fromDBName("enwiki")
        for item in pg.WikidataSPARQLPageGenerator(
            query, site=self.bot.site, result_type=list
        ):
            item.get()
            log.debug(
                "set_dbname2item " + item.title() + " " + str(item.labels.get("en"))
            )
            if "enwiki" not in item.sitelinks:
                log.debug(
                    "ignore " + item.title() + " because it does not link to enwiki"
                )
                continue
            p = pywikibot.Page(enwiki, item.sitelinks["enwiki"].title)
            self.license2item[p.title()] = item

    def set_license2item(self):
        format_args = {
            "subclass_of": self.P_subclass_of,
            "instance_of": self.P_instance_of,
            "open_source": self.Q_open_source_license.getID(),
            "free_software": self.Q_free_software_license.getID(),
            "licenses": "",
        }
        if self.args.license:
            licenses = []
            for license in self.args.license:
                licenses.append("STR(?label) = '" + license + "'")
            licenses = (
                "?item rdfs:label ?label FILTER((" + ") || (".join(licenses) + "))"
            )
            format_args["licenses"] = licenses
        query = """
            SELECT DISTINCT ?item WHERE {{
              {{
                ?item wdt:{instance_of}?/wdt:{subclass_of}* wd:{open_source}.
              }} Union {{
                ?item wdt:{instance_of}?/wdt:{subclass_of}* wd:{free_software}.
              }}
              {licenses}
        }}
        """.format(
            **format_args
        )
        log.debug("set_license2item " + query)
        self.license2item = {}
        enwiki = pywikibot.site.APISite.fromDBName("enwiki")
        for item in pg.WikidataSPARQLPageGenerator(
            query, site=self.bot.site, result_type=list
        ):
            item.get()
            log.debug(
                "set_license2item " + item.title() + " " + str(item.labels.get("en"))
            )
            if "enwiki" not in item.sitelinks:
                log.debug(
                    "set_license2item ignore "
                    + item.title()
                    + " because it does not link to enwiki"
                )
                continue
            p = pywikibot.Page(enwiki, item.sitelinks["enwiki"].title)
            self.license2item[p.title()] = item

    def template_parse_license(self, license, lang):
        free_software_licenses = self.get_names(lang)
        results = set()
        for name in re.findall("\[\[([^|\]]+?)\]\]", license) + re.findall(
            "\[\[([^|\]]+?)\|[^\]]*\]\]", license
        ):
            name = re.sub("#.*", "", name)
            log.debug("template_parse_license: " + name)
            if name in free_software_licenses:
                results.add(self.get_item(name, lang))
        return list(results)

    def fixup(self, item):
        item.get()

        if self.P_license in item.claims:
            return ["exists"]

        lang2field = {
            "ca": "Llicència",
            "en": "License",
            "ja": "license",
            "ml": "license",
            "ru": "license",
            "zh": "license",
        }
        lang2template = {
            "ca": "Caixa Programari",
            "es": "Ficha de software",
            "en": "Infobox",
            "it": "Software",
            "pt": "Info/Software",
            "ru": "Карточка программы",
            "*": "Infobox",
        }
        lang2value = {}
        for lang, license in self.get_template_field(
            item, lang2field, lang2template
        ).items():
            lang2value[lang] = self.template_parse_license(license, lang)
        if len(lang2value) == 0:
            return ["nothing"]
        self.debug(item, "fixup " + str(lang2value))
        values = list(lang2value.values())
        # if one wikipedia disagrees with the others, do nothing
        if values.count(values[0]) != len(values):
            self.error(
                item,
                "inconsistent license information between wikipedia" + str(lang2value),
            )
            return ["inconsistent"]
        status = []
        for license in lang2value[list(lang2value.keys())[0]]:
            license.get()
            langs = list(lang2value.keys())
            self.info(
                item, "ADD license " + license.labels["en"] + " from " + str(langs)
            )
            status.append(license.labels["en"])
            claim = pywikibot.Claim(self.bot.site, self.P_license, 0)
            claim.setTarget(license)
            if not self.args.dry_run:
                item.addClaim(claim)
            for lang in langs:
                imported = pywikibot.Claim(
                    self.bot.site, self.P_imported_from, is_reference=True
                )
                imported.setTarget(self.get_sitelink_item(lang + "wiki"))
                if not self.args.dry_run:
                    claim.addSource(imported)
            self.set_retrieved(item, claim)
        return status



# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import subprocess

log = logging.getLogger(__name__)


def sh_bool(command):
    try:
        sh(command)
        return True
    except subprocess.CalledProcessError:
        return False


def sh(command, input=None):
    log.debug(":sh: " + command)
    if input is None:
        stdin = None
    else:
        stdin = subprocess.PIPE
    proc = subprocess.Popen(
        args=command,
        stdin=stdin,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        bufsize=1,
    )
    if stdin is not None:
        proc.stdin.write(input.encode("ascii", "ignore"))
        proc.stdin.close()
    lines = []
    with proc.stdout:
        for line in iter(proc.stdout.readline, b""):
            line = line.decode("utf-8", "ignore")
            lines.append(line)
            log.debug(line.strip().encode("ascii", "ignore"))
    if proc.wait() != 0:
        raise subprocess.CalledProcessError(returncode=proc.returncode, cmd=command)
    return "".join(lines)



# -*- mode: python; coding: utf-8 -*-
# SPDX-License-Identifier: GPL-3.0-or-later

import subprocess

import pytest

from f3wikidatabot import util


class TestUtil(object):
    def test_sh(self):
        assert "A" == util.sh("echo -n A")
        with pytest.raises(Exception) as excinfo:
            util.sh("exit 123")
        assert excinfo.value.returncode == 123

    def test_sh_progress(self, caplog):
        util.sh("echo AB ; sleep 5 ; echo C")
        records = caplog.records
        assert ":sh: " in records[0].message
        assert "b'AB'" == records[1].message
        assert "b'C'" == records[2].message

    def test_sh_input(self, caplog):
        assert "abc" == util.sh("cat", "abc")

    def test_sh_fail(self, caplog):
        with pytest.raises(subprocess.CalledProcessError) as excinfo:
            util.sh("/bin/echo -n AB ; /bin/echo C ; exit 111")
        assert excinfo.value.returncode == 111

        for record in caplog.records:
            if record.levelname == "ERROR":
                assert "replay full" in record.message or "ABC\n" == record.message

    def test_sh_bool__returns_true_on_exit_code_zero(self):
        assert True is util.sh_bool("true")

    def test_sh_bool__returns_false_on_non_zero_exit_code(self):
        assert False is util.sh_bool("false")

    def test_sh__returns_empty_string(self):
        assert "" == util.sh("true")

    def test_sh__handles_utf8(self):
        assert "€" == util.sh("echo -n €")



# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import logging
import textwrap
import time

import pywikibot
from pywikibot import pagegenerators as pg

from f3wikidatabot import license
from f3wikidatabot import features
from f3wikidatabot.plugin import Plugin

logging.basicConfig(format="%(asctime)s %(levelname)s %(message)s")
log = logging.getLogger(__name__)

plugins = [
    license.License,
    features.Features,
]

name2plugin = dict([(p.__name__, p) for p in plugins])


class Bot(object):
    def __init__(self, args):
        self.args = args
        logging.getLogger("f3wikidatabot").setLevel(self.args.verbose)
        self.site = pywikibot.Site(
            code="wikidata" if not self.args.test else "test",
            fam="wikidata",
            user=self.args.user,
        )
        if self.args.test:
            self.site.throttle.setDelays(writedelay=0)
        if self.args.test:
            self.wikidata_site = pywikibot.Site(code="wikidata", fam="wikidata")
        else:
            self.wikidata_site = None
        self.plugins = []
        for name in self.args.plugin or name2plugin.keys():
            plugin = name2plugin[name]
            self.plugins.append(plugin(self, args))

    @staticmethod
    def get_parser():
        filters = []
        available_plugins = []
        for plugin in plugins:
            filters.extend(plugin.filter_names())
            available_plugins.append(plugin.__name__)
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_const",
            const=logging.DEBUG,
            default=logging.INFO,
        )
        parser.add_argument(
            "--dry-run", action="store_true", default=None, help="no side effect"
        )
        parser.add_argument(
            "--test",
            action="store_true",
            default=None,
            help="use test.wikidata.org instead of wikidata.org",
        )
        parser.add_argument("--user", default=None, help="wikidata user name")
        parser.add_argument(
            "--plugin",
            default=[],
            choices=available_plugins,
            action="append",
            help="use this plugin instead of all of them (can be repeated)",
        )
        parser.add_argument(
            "--show",
            action="store_true",
            default=None,
            help="show items",
        )
        select = parser.add_mutually_exclusive_group()
        select.add_argument(
            "--filter",
            default="",
            choices=filters,
            help="filter with a pre-defined query",
        )
        select.add_argument(
            "--item",
            default=[],
            action="append",
            help="work on this QID (can be repeated)",
        )
        return parser

    @staticmethod
    def factory(argv):
        parents = [
            Bot.get_parser(),
            Plugin.get_parser(),
        ]
        for plugin in plugins:
            parents.append(plugin.get_parser())
        parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=textwrap.dedent(
                """\
            A command-line toolbox for the wikidata Forges project.
            """
            ),
            parents=parents,
        )
        return Bot(parser.parse_args(argv))

    def run(self):
        if len(self.args.item) > 0:
            results = self.run_items()
        else:
            results = self.run_query()
        if self.args.show:
            for result in results:
                print(result.to_json())

    def run_items(self):
        results = []
        for item in self.args.item:
            item = pywikibot.ItemPage(self.site, item, 0)
            for plugin in self.plugins:
                results.append(plugin.run_catch(item))
        return results

    def run_query(self):
        for plugin in self.plugins:
            query = plugin.get_query(self.args.filter)
            if query is not None:
                break
        if query is None:
            query = Plugin(self, self.args).get_query(self.args.filter)
        query = query + " # " + str(time.time())
        log.debug("running query " + query)
        results = []
        for item in pg.WikidataSPARQLPageGenerator(
            query, site=self.site, result_type=list
        ):
            for plugin in self.plugins:
                results.append(plugin.run_catch(item))
        return results



# SPDX-License-Identifier: GPL-3.0-or-later

import random
import string

import pywikibot
from pywikibot.login import ClientLoginManager


class WikidataHelper(object):
    def login(self):
        site = pywikibot.Site("test", "wikidata", "F3BotCI")
        ClientLoginManager(
            site=site, user="F3BotCI", password="quars3Knafs5"
        ).login_to_site()
        return site

    @staticmethod
    def random_name():
        return "".join(random.choice(string.ascii_lowercase) for _ in range(16))



# -*- mode: python; coding: utf-8 -*-
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from unittest import mock
import pywikibot

from f3wikidatabot.bot import Bot
from f3wikidatabot.license import License
from f3wikidatabot.tests.wikidata import WikidataHelper

log = logging.getLogger("f3wikidatabot")


class TestLicense(object):
    def setup_class(self):
        self.site = WikidataHelper().login()

    def teardown_class(self):
        self.site.logout()

    def setup(self):
        self.gpl = "GNU General Public License"
        self.mit = "MIT license"
        self.args = [
            "--license",
            self.gpl,
            "--license",
            self.mit,
        ]

    def test_get_item(self):
        bot = Bot.factory(["--verbose"] + self.args)
        license = License(bot, bot.args)
        redirect = "GPL"
        license.get_names("en")
        canonical_item = license.get_item(self.gpl, "en")
        assert canonical_item == license.get_item(redirect, "en")

        gpl_fr = "Licence publique générale GNU"
        names_fr = license.get_names("fr")
        assert gpl_fr in names_fr
        assert canonical_item == license.get_item(gpl_fr, "fr")

    def test_get_names(self):
        bot = Bot.factory(["--verbose"] + self.args)
        license = License(bot, bot.args)
        redirect = "GPL"
        names = license.get_names("en")
        assert self.gpl in names
        assert redirect in names

        canonical_fr = "Licence publique générale GNU"
        names = license.get_names("fr")
        assert canonical_fr in names
        assert self.gpl in names

    def test_template_parse_license(self):
        bot = Bot.factory(["--verbose"] + self.args)
        license = License(bot, bot.args)
        found = license.template_parse_license(
            "[[GNU GPL#v2]] [[MIT/X11 license|]]", "en"
        )
        assert 2 == len(found)
        for item in found:
            item.get()
            license.debug(item, "FOUND")
            assert item.labels["en"] in (self.gpl, self.mit)

    @mock.patch("f3wikidatabot.license.License.set_license2item")
    @mock.patch("f3wikidatabot.plugin.Plugin.get_sitelink_item")
    def test_fixup(self, m_get_sitelink_item, m_set_license2item):
        bot = Bot.factory(
            [
                "--verbose",
                "--test",
                "--user=F3BotCI",
            ]
        )
        license = License(bot, bot.args)
        gpl = license.Q_GNU_General_Public_License
        gpl.get()
        found = False
        if gpl.claims:
            for claim in gpl.claims.get(license.P_subclass_of, []):
                if claim.type != "wikibase-item":
                    continue
                if claim.getTarget().getID() == license.Q_free_software_license.getID():
                    found = True
                    break
        if not found:
            subclass_of = pywikibot.Claim(license.bot.site, license.P_subclass_of, 0)
            subclass_of.setTarget(license.Q_free_software_license)
            gpl.addClaim(subclass_of)
        gpl.setSitelink({"site": "enwiki", "title": self.gpl})
        gpl.get(force=True)

        emacs = license.Q_GNU_Emacs
        emacs.get()
        if emacs.claims:
            licenses = emacs.claims.get(license.P_license, [])
            if licenses:
                emacs.removeClaims(licenses)
                emacs.get(force=True)

        def set_license2item():
            license.license2item = {self.gpl: license.Q_GNU_General_Public_License}

        m_set_license2item.side_effect = set_license2item

        def get_sitelink_item(dbname):
            if dbname == "enwiki":
                return license.Q_English_Wikipedia
            elif dbname == "frwiki":
                return license.Q_French_Wikipedia
            else:
                assert 0, "unexpected " + dbname

        m_get_sitelink_item.side_effect = get_sitelink_item
        emacs.removeSitelinks(["enwiki"])
        emacs.removeSitelinks(["frwiki"])
        emacs.get(force=True)
        assert ["nothing"] == license.fixup(emacs)

        emacs.setSitelink({"site": "enwiki", "title": "GNU Emacs"})
        emacs.setSitelink({"site": "frwiki", "title": "GNU Emacs"})
        emacs.get(force=True)
        assert [self.gpl] == license.fixup(emacs)


# Local Variables:
# compile-command: "cd .. ; tox -e py3 tests/test_license.py"
# End:



# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import logging
import re

import pywikibot
from pywikibot import pagegenerators as pg

from f3wikidatabot.forge import Forge
from f3wikidatabot import plugin

log = logging.getLogger(__name__)


class Features(plugin.Plugin):
    def __init__(self, *args):
        super(Features, self).__init__(*args)

    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser(add_help=False)
        return parser

    @staticmethod
    def filter_names():
        return []

    def get_query(self, filter):
        format_args = {
            "instance_of": self.P_instance_of,
            "discontinued_date": self.P_discontinued_date,

            "website": self.Q_website.getID(),
            "forge": self.Q_forge.getID(),
        }
        query = """
         SELECT ?item
         WHERE
         {{
           {{
             ?item wdt:{instance_of} wd:{forge}.
             MINUS {{
                 ?item p:{discontinued_date} ?ignore0.
             }}
           }}
           FILTER NOT EXISTS {{
             ?item wdt:{instance_of}+ wd:{website}.
           }} # exclude online services running a forge, only keep forge software
         }}
        """.format(
            **format_args
        )
        return query

    def run(self, item):
        self.fixup(item)
        self.verify(item)

    def feature_items(self):
        #
        # For an authoritative list of feature items, see
        # https://www.wikidata.org/wiki/Wikidata:WikiProject_Informatics/Forges#Properties
        #
        return [
            self.Q_repository.getID(),
            self.Q_issue_tracking_system.getID(),
            self.Q_continuous_integration_software.getID(),
            self.Q_code_reviewing_software.getID(),
            self.Q_collaborative_wiki_software.getID(),
            self.Q_electronic_mailing_list_manager.getID(),
            self.Q_software_repository.getID(),
            self.Q_pull_request.getID(),
        ]

    def show(self, item):
        item.get()
        forge = Forge(id=item.getID(),
                      label=item.labels.get("en", "no label"))
        features = self.feature_items()
        for instance_of in item.claims.get(self.P_instance_of, []):
            target = instance_of.getTarget()
            if target.getID() in features:
                forge.features.append(target.labels.get("en", "no label"))
            if target.getID() == self.Q_proprietary_software.getID():
                forge.proprietary = True
        for based_on in item.claims.get(self.P_based_on, []):
            target = based_on.getTarget()
            forge.based_on.append(target.labels.get("en", "no label"))
        self.debug(item, f"show {forge}")
        return forge

    def verify(self, item):
        item.get()

        self.debug(item, "verify")
        status = {}
        for claim in item.claims.get(self.P_instance_of, []):
            pass
        return status

    def fixup(self, item):
        pass



# -*- mode: python; coding: utf-8 -*-
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import date
import logging

import pytest
import pywikibot
import time

from f3wikidatabot.bot import Bot
from f3wikidatabot.plugin import Plugin
from f3wikidatabot.tests.wikidata import WikidataHelper

log = logging.getLogger("f3wikidatabot")


class TestPlugin(object):
    def setup_class(self):
        self.site = WikidataHelper().login()

    def teardown_class(self):
        self.site.logout()

    def test_lookup_item(self):
        bot = Bot.factory(
            [
                "--test",
                "--user=F3BotCI",
                "--verbose",
            ]
        )
        plugin = Plugin(bot, bot.args)
        assert 0 == len(plugin.bot.entities["item"])

        git = plugin.Q_git
        assert 1 == len(plugin.bot.entities["item"])
        assert git == plugin.Q_git
        assert plugin.Q_Concurrent_Versions_System
        assert 2 == len(plugin.bot.entities["item"])

    def test_create_entity(self):
        bot = Bot.factory(
            [
                "--test",
                "--user=F3BotCI",
                "--verbose",
            ]
        )
        plugin = Plugin(bot, bot.args)
        name = "Q_" + WikidataHelper.random_name()
        log.debug(f"looking for {name}")
        item = plugin.__getattribute__(name)
        assert 1 == len(plugin.bot.entities["item"])

        plugin.clear_entity_label(item.getID())
        assert 0 == len(plugin.bot.entities["item"])

        item = plugin.__getattribute__(name)
        assert 1 == len(plugin.bot.entities["item"])

        log.debug(f"moving to property2datatype")

        #
        # These properties must exist in wikidata.org and are used to create the same
        # property in test.wikidata.org
        #
        property2datatype = {
            "P_source_code_repository_URL": "url",
            "P_chemical_formula": "string",
            "P_based_on": "wikibase-item",
        }

        wikidata_bot = Bot.factory(
            [
                "--user=F3BotCI",
                "--verbose",
            ]
        )
        wikidata_plugin = Plugin(wikidata_bot, wikidata_bot.args)
        for attr, datatype in property2datatype.items():
            log.debug(f"looking for {name} datatype {datatype}")
            plugin.reset_cache()
            #
            # because plugin uses test.wikidata.org (--test), it will
            # create the property if it does not exist already. It would
            # fail if it was wikidata.org because this is not allowed.
            #
            property = plugin.__getattribute__(attr)
            assert 1 == len(plugin.bot.entities["property"])
            #
            # clear the property label so it can no longer be found
            #
            plugin.clear_entity_label(property)
            assert 0 == len(plugin.bot.entities["property"])
            #
            # retry because the change is asynchronous and the property
            # may still have the label that allows to find it for a short
            # while
            #
            for _ in range(120):
                if plugin.lookup_entity(attr, type="property") is None:
                    break
                log.debug(f"retry lookup_entity: {attr}")
            #
            # this will fail to find the property and re-create it
            #
            property = plugin.__getattribute__(attr)
            assert 1 == len(plugin.bot.entities["property"])

            #
            # verify that the property created in test.wikidata.org has the
            # same type as the property that exists in wikidata.org
            #
            new_content = plugin.bot.site.loadcontent({"ids": property}, "datatype")
            wikidata_property = wikidata_plugin.__getattribute__(attr)
            wikidata_content = wikidata_plugin.bot.site.loadcontent(
                {"ids": wikidata_property}, "datatype"
            )
            assert (
                wikidata_content[wikidata_property]["datatype"]
                == new_content[property]["datatype"]
            ), attr
            assert datatype == wikidata_content[wikidata_property]["datatype"], attr

    def test_set_retrieved(self):
        bot = Bot.factory(
            [
                "--test",
                "--user=F3BotCI",
            ]
        )
        plugin = Plugin(bot, bot.args)
        item = plugin.__getattribute__("Q_" + WikidataHelper.random_name())
        claim = pywikibot.Claim(plugin.bot.site, plugin.P_source_code_repository_URL, 0)
        claim.setTarget("http://repo.com/some")
        item.addClaim(claim)
        plugin.set_retrieved(item, claim)
        assert plugin.need_verification(claim) is False

        plugin.set_retrieved(item, claim, date(1965, 11, 2))
        assert plugin.need_verification(claim) is True

        plugin.clear_entity_label(item.getID())

    def test_search_entity(self):
        bot = Bot.factory(
            [
                "--test",
                "--user=F3BotCI",
                "--verbose",
            ]
        )
        plugin = Plugin(bot, bot.args)
        # ensure space, - and _ are accepted
        name = WikidataHelper.random_name() + "-some thing_else"

        #
        # Insert two items that have the same label. Then search for
        # the label and check that it throws an exception. Having two
        # items with the same label is a source of confusion and needs
        # manual fixing.
        #
        first = plugin.bot.site.editEntity({"new": "item"}, {})
        first_entity = plugin.set_entity_label(first["entity"]["id"], name)
        second = plugin.bot.site.editEntity({"new": "item"}, {})
        second_entity = plugin.set_entity_label(second["entity"]["id"], name)

        assert first["entity"]["id"] != second["entity"]["id"]

        with pytest.raises(ValueError) as e:
            for _ in range(120):
                #
                # It takes time for test.wikidata.org to return
                # **both** items in the search (wbsearchentities API
                # endpoint), wait for it.
                #
                plugin.search_entity(plugin.bot.site, name, type="item")
                time.sleep(5)
        assert "found multiple items" in str(e.value)

        #
        # Special case implemented by plugin.search_entity. If an item
        # is an instance of a disambiguation page, it is not returned
        # in the search results.
        #
        claim = pywikibot.Claim(plugin.bot.site, plugin.P_instance_of, 0)
        claim.setTarget(plugin.Q_Wikimedia_disambiguation_page)
        first_entity.addClaim(claim)

        found = plugin.search_entity(bot.site, name, type="item")
        assert found.getID() == second_entity.getID()

        third = plugin.bot.site.editEntity({"new": "item"}, {})
        third_entity = plugin.set_entity_label(third["entity"]["id"], name)

        with pytest.raises(ValueError) as e:
            for _ in range(120):
                #
                # It takes time for test.wikidata.org to return
                # **both** items in the search (wbsearchentities API
                # endpoint), wait for it.
                #
                plugin.search_entity(plugin.bot.site, name, type="item")
                time.sleep(5)
        assert "found multiple items" in str(e.value)

        Plugin.authoritative["test"][name] = second_entity.getID()
        found = plugin.search_entity(plugin.bot.site, name, type="item")
        assert found.getID() == second_entity.getID()

    def test_get_template_field(self):
        bot = Bot.factory(["--verbose"])
        plugin = Plugin(bot, bot.args)
        item = plugin.Q_GNU_Emacs
        expected = {
            "simple": "ignore",
            "nl": "ignore",
            "en": "ignore",
        }
        item.get()
        lang2field = {"en": "License"}
        lang2pattern = {"*": "Infobox"}
        actual = plugin.get_template_field(item, lang2field, lang2pattern)
        assert actual.keys() == expected.keys()

    def test_translate_title(self):
        bot = Bot.factory(["--verbose"])
        plugin = Plugin(bot, bot.args)
        assert "Licence" == plugin.translate_title("License", "fr")
        assert plugin.translate_title("License", "??") is None

    def test_get_redirects(self):
        bot = Bot.factory(["--verbose"])
        plugin = Plugin(bot, bot.args)
        titles = plugin.get_redirects("GNU General Public License", "en")
        assert "GPL" in titles

    def test_get_sitelink_item(self):
        bot = Bot.factory(["--verbose"])
        plugin = Plugin(bot, bot.args)
        enwiki = plugin.get_sitelink_item("enwiki")
        assert "English Wikipedia" == enwiki.labels["en"]

        frwiki = plugin.get_sitelink_item("frwiki")
        assert "French Wikipedia" == frwiki.labels["en"]



# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass, field
from dataclasses_json import config, dataclass_json

@dataclass_json
@dataclass
class Forge:
    id: str
    label: str
    features: list[str] = field(default_factory=list)
    based_on: list[str] = field(default_factory=list)
    proprietary: bool = False



# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import logging
import re
from datetime import datetime, timedelta

import pywikibot
import requests
from pywikibot import pagegenerators as pg

log = logging.getLogger(__name__)


class Plugin(object):
    def __init__(self, bot, args):
        self.args = args
        self.bot = bot
        self.reset_cache()
        self.title_translation = {}
        self.dbname2item = {}

    @staticmethod
    def get_parser():
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument(
            "--verification-delay",
            type=int,
            default=30,
            help="days to wait before verifying a claim again",
        )
        return parser

    def get_query(self, filter):
        query = """
        SELECT DISTINCT ?item WHERE {{
          ?item wdt:{source_code_repository_url} ?url.
        }} ORDER BY ?item
        """.format(
            source_code_repository_url=self.P_source_code_repository_URL
        )
        return query

    def debug(self, item, message):
        self.log(log.debug, item, message)

    def info(self, item, message):
        self.log(log.info, item, message)

    def error(self, item, message):
        self.log(log.error, item, message)

    def log(self, fun, item, message):
        label = item.labels.get("en", "no label")
        fun(
            "http://wikidata.org/wiki/"
            + item.getID()
            + " "
            + label
            + " "
            + self.__class__.__name__
            + " "
            + message
        )

    def run_catch(self, item):
        try:
            if self.args.show:
                return self.show(item)
            else:
                return self.run(item)
        except:
            self.error(item, "failed with an exception")
            raise

    def reset_cache(self):
        self.bot.entities = {
            "property": {},
            "item": {},
        }

    def lookup_entity(self, name, **kwargs):
        type = kwargs["type"]
        found = self.bot.entities[type].get(name)
        if found:
            return found
        found = self.search_entity(self.bot.site, name, **kwargs)
        if found:
            if type == "property":
                found = found["id"]
            self.bot.entities[type][name] = found
        return found

    #
    # Hardcode the desired wikidata item when there are
    # multiple items with the same english label and no
    # trivial way to disambiguate them.
    #
    authoritative = {
        "wikidata": {
            "repository": "Q3133368",
            "issue_tracking_system": "Q1480561",
            "continuous_integration_software": "Q16947796",
            "code_reviewing_software": "Q16920237",
            "collaborative_wiki_software": "Q6686945",
            "electronic_mailing_list_manager": "Q63067479",
            "software_repository": "Q1334294",
            "pull_request": "Q68712963",
            "forge": "Q3077240",
            "website": "Q35127",
        },
        "test": {},
    }

    @staticmethod
    def normalize_name(name):
        return re.sub("[-_]", " ", name)

    def search_entity(self, site, name, **kwargs):
        if name in Plugin.authoritative[site.code]:
            candidate = pywikibot.ItemPage(
                site, Plugin.authoritative[site.code][name], 0
            )
            if candidate.get()["labels"]["en"] == name:
                return candidate
        candidates = []
        for p in site.search_entities(name, "en", **kwargs):
            log.debug("looking for entity `" + name + "`, found " + str(p))
            if "label" in p and Plugin.normalize_name(
                p["label"]
            ) == Plugin.normalize_name(name):
                if kwargs["type"] == "property":
                    candidates.append(p)
                else:
                    candidates.append(pywikibot.ItemPage(site, p["id"], 0))
        if len(candidates) == 0:
            return None
        elif len(candidates) > 1 and kwargs["type"] == "item":
            found = []
            for candidate in candidates:
                item = candidate.get()
                ok = True
                for instance_of in item["claims"].get(self.P_instance_of, []):
                    if instance_of.getTarget() == self.Q_Wikimedia_disambiguation_page:
                        log.debug(
                            "ignore disambiguation page "
                            + candidate.getID()
                            + " for "
                            + name
                        )
                        ok = False
                        break
                if ok:
                    found.append(candidate)
            if len(found) != 1:
                raise ValueError("found multiple items for " + name + " " + str(found))
            return found[0]
        else:
            return candidates[0]

    lookup_item = lookup_entity

    def lookup_property(self, name):
        return self.lookup_entity(self.bot.site, name, type="property")

    def create_entity(self, type, name):
        found = self.search_entity(self.bot.wikidata_site, name, type=type)
        entity = {
            "labels": {
                "en": {
                    "language": "en",
                    "value": name,
                }
            },
        }
        if type == "property":
            assert found, type + " " + name + " must exist in wikidata"
            id = found["id"]
            found = self.bot.wikidata_site.loadcontent({"ids": id}, "datatype")
            assert found, "datatype of " + id + " " + name + " is not found"
            entity["datatype"] = found[id]["datatype"]
        log.debug("create " + type + " " + str(entity))
        self.bot.site.editEntity({"new": type}, entity)

    def clear_entity_label(self, id):
        self.set_entity_label(id, "")

    def set_entity_label(self, id, label):
        data = {
            "labels": {
                "en": {
                    "language": "en",
                    "value": label,
                }
            }
        }
        log.debug("set " + id + " label to '" + label + "'")
        self.bot.site.editEntity({"id": id}, data)
        while True:
            if id.startswith("P"):
                entity = pywikibot.PropertyPage(self.bot.site, id)
            else:
                entity = pywikibot.ItemPage(self.bot.site, id, 0)
            entity.get(force=True)
            if label == "" and entity.labels.get("en") is None:
                break
            if label != "" and label == entity.labels.get("en"):
                break
        self.reset_cache()
        return entity

    def __getattribute__(self, name):
        if name.startswith("P_"):
            type = "property"
        elif name.startswith("Q_"):
            type = "item"
        else:
            return super(Plugin, self).__getattribute__(name)
        label = " ".join(name.split("_")[1:])
        found = self.lookup_entity(label, type=type)
        if not found:
            if self.args.test:
                self.create_entity(type, label)
                for _ in range(120):
                    found = self.lookup_entity(label, type=type)
                    if found is not None:
                        break
            else:
                raise ValueError("found no items for " + name)
        return found

    def get_source(self, claim, id):
        for source in claim.getSources():
            if id in source:
                return source[id]
        return None

    def need_verification(self, claim):
        previous = self.get_source(claim, self.P_retrieved)
        if previous:
            now = datetime.utcnow()
            previous = previous[0].getTarget()
            previous = datetime(
                year=previous.year, month=previous.month, day=previous.day
            )
            return now - previous >= timedelta(days=self.args.verification_delay)
        else:
            return True

    def set_retrieved(self, item, claim, now=datetime.utcnow()):
        when = pywikibot.WbTime(now.year, now.month, now.day)
        retrieved = self.get_source(claim, self.P_retrieved)
        if retrieved:
            self.debug(item, "updating retrieved")
            retrieved[0].setTarget(when)
            if not self.args.dry_run:
                self.bot.site.save_claim(claim)
        else:
            self.debug(item, "setting retrieved")
            retrieved = pywikibot.Claim(
                self.bot.site, self.P_retrieved, is_reference=True
            )
            retrieved.setTarget(when)
            if not self.args.dry_run:
                claim.addSource(retrieved)

    def http_get(self, url):
        try:
            #
            # although head() would be more light weight, some
            # servers do not respond to it. For instance
            # https://src.openvz.org/projects/OVZ/ returned 405
            #
            # The user agent is required for some servers. For
            # instance http://marabunta.laotracara.com/descargas/
            # returns 406 if no User-Agent header is set.
            #
            r = requests.get(url, headers={"User-Agent": "f3wikidatabot"}, verify=False)
            log.debug("GET " + url + " status " + str(r.status_code))
            if r.status_code != requests.codes.ok:
                log.debug("GET " + url + " " + r.text)
            if r.status_code == requests.codes.ok:
                return r
            else:
                return None
        except Exception as e:
            log.debug("GET failed with " + str(e))
            return None

    def get_template_field(self, item, lang2field, lang2pattern):
        """For a given `item`, look into all wikipedia pages linked to it.
        Look for known templates in the page in accordance to `lang2pattern`.
        If a known template is found, look for a field as listed in `lang2field`
        and return the associated value."""
        lang2value = {}
        for dbname in item.sitelinks.keys():
            site = pywikibot.site.APISite.fromDBName(dbname)
            pattern = lang2pattern.get(site.code, lang2pattern["*"])
            p = pywikibot.Page(site, item.sitelinks[dbname].title)
            for template, pairs in p.templatesWithParams():
                self.debug(item, site.code + " template " + template.title())
                if pattern in template.title():
                    for pair in pairs:
                        found = pair.split("=", 1)
                        if len(found) == 1:
                            continue
                        (name, value) = found
                        if site.code in lang2field:
                            translated = lang2field[site.code]
                        elif "en" in lang2field:
                            translated = self.translate_title(
                                lang2field["en"], site.code
                            )
                        else:
                            translated = None
                        self.debug(
                            item,
                            site.code
                            + " compare "
                            + str(translated).lower()
                            + " and "
                            + name.lower(),
                        )
                        if value and translated and name.lower() == translated.lower():
                            lang2value[site.code] = value
        self.debug(item, "get_template_field " + str(lang2value))
        return lang2value

    def translate_title(self, title, lang):
        if title not in self.title_translation:
            site = pywikibot.site.APISite.fromDBName("enwiki")
            translation = {"en": title}
            p = pywikibot.Page(site, title)
            for l in p.langlinks():
                # License (juridique): the (...) is for disambiguation
                translation[l.site.code] = re.sub("\s*\([^)]*\)", "", l.title)
            self.title_translation[title] = translation
        return self.title_translation[title].get(lang)

    def get_redirects(self, title, lang):
        site = pywikibot.site.APISite.fromDBName(lang + "wiki")
        p = pywikibot.Page(site, title)
        r = [
            r.title()
            for r in p.getReferences(
                follow_redirects=False,
                with_template_inclusion=False,
                filter_redirects=True,
                total=5000,
            )
        ]
        log.debug("get_redirects " + title + " " + lang + " " + str(r))
        return r

    def get_sitelink_item(self, dbname):
        if dbname not in self.dbname2item:
            query = """
            SELECT DISTINCT ?item WHERE {{
              ?item wdt:{Wikimedia_database_name} '{dbname}'.
            }}
            """.format(
                Wikimedia_database_name=self.P_Wikimedia_database_name,
                dbname=dbname,
            )
            for item in pg.WikidataSPARQLPageGenerator(query, site=self.bot.site):
                item.get()
                assert (
                    dbname == item.claims[self.P_Wikimedia_database_name][0].getTarget()
                )
                self.dbname2item[dbname] = item
        return self.dbname2item[dbname]



# -*- mode: python; coding: utf-8 -*-
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from unittest import mock
import pywikibot

from f3wikidatabot.bot import Bot
from f3wikidatabot.features import Features
from f3wikidatabot.tests.wikidata import WikidataHelper

log = logging.getLogger("f3wikidatabot")


class TestFeatures(object):
    def setup_class(self):
        self.site = WikidataHelper().login()

    def teardown_class(self):
        self.site.logout()

    def setup(self):
        pass

    def test_query(self):
        bot = Bot.factory(["--verbose"])
        features = Features(bot, bot.args)
        query = features.get_query([])
        assert "P31" in query

    def test_show(self):
        bot = Bot.factory(["--verbose", "--show", "--plugin", "Features"])
        forgejo = "Forgejo"
        heptapod = "Heptapod"
        gitlab = "GitLab EE"
        found = []
        for forge in bot.run_query():
            if forge.label == forgejo:
                found.append(forge.label)
                assert "software repository" in forge.features
                assert forge.proprietary == False
            elif forge.label == heptapod:
                found.append(forge.label)
                assert "repository" in forge.features
                assert forge.proprietary == False
            elif forge.label == gitlab:
                found.append(forge.label)
                assert "GitLab FOSS" in forge.based_on
                assert forge.proprietary == True
        assert sorted(found) == sorted([forgejo, heptapod, gitlab])
